/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <libintl.h>

#include "sambascanner.h"
#include "String.h"
#include "Parser.h"
#include "IP_Range.h"
#include "CommandLine.h"

extern struct Options_Structure Options;

void Display_Help(char **argv) {
	fprintf(stderr, _("Syntax: %s [Options]\n"), argv[0]);
	fprintf(stderr, "\n");
	fprintf(stderr, _("-t, --time       Requires a float argument which states by which factor the\n                 internal timeouts should be multiplied with.\n"));
	fprintf(stderr, _("-w, --workgroup  Manually sets the workgroup to the given argument.\n"));
	fprintf(stderr, _("-d, --debug      Set the debug level from 0 (silent) to 6 (full debug),\n                 default level is 1.\n"));
	fprintf(stderr, _("-s, --smbclient  Set the file and path where smbclient can be found (default is\n                 %s).\n"), DEFAULT_SMBCLIENT);
	fprintf(stderr, _("-n, --nmblookup  Set the file and path where nmblookup can be found (default is\n                 %s).\n"), DEFAULT_NMBLOOKUP);
	fprintf(stderr, _("-l, --smblister  Set the file and path where smblister can be found (default is\n                 %s).\n"), DEFAULT_SMBLISTER);
	fprintf(stderr, _("-c, --computer   Specify a computer on the network.\n"));
	fprintf(stderr, _("-o, --outfile    Set the output file (default 'Filelist.txt').\n"));
	fprintf(stderr, _("-i, --ip         Specify an IP or IP range to scan (either 'x.x.x.x',\n                 'x.x.x.x/n' or 'a-b.a-b.a-b.a-b').\n"));
	fprintf(stderr, _("-a, --no-passwd  Don't try some default passwords for shares which won't open up\n                 immediately.\n"));
	fprintf(stderr, _("-r, --threadcnt  Specifies how many threads should be used (default 30).\n"));
	fprintf(stderr, _("-p, --preserve   Doesn't delete tempfiles after work is done.\n"));
	fprintf(stderr, _("-u, --userpass   Specify a username and passwort which smblister will use during\n                 the scanning process. Has to be supplied in the form\n                 'Username,Password'.\n"));
	fprintf(stderr, _("-f, --nofilter   Do not filter out 'print$' shares. These are not scanned by\n                 default.\n"));
	fprintf(stderr, _("-h, --help       Display this help.\n"));
	fprintf(stderr, "\n");
	fprintf(stderr, _("Examples: %s -w MYGROUP -t 2.5\n"), argv[0]);
	fprintf(stderr, _("          Multiplies all timeouts by 2.5 and sets the workgroup to 'MYGROUP'.\n"));
	fprintf(stderr, "\n");
	fprintf(stderr, _("          %s -w MATRIX -c NEO\n"), argv[0]);
	fprintf(stderr, _("          Select workgroup 'MYGROUP' and specify known computer 'NEO'.\n"));
	fprintf(stderr, "\n");
	fprintf(stderr, _("          %s -i 172.17.5.0/24\n"), argv[0]);
	fprintf(stderr, _("          Scans computers 172.17.5.0 to 172.17.5.255.\n"));
	fprintf(stderr, "\n");
	fprintf(stderr, _("          %s -i 172.17.5-6.1-127\n"), argv[0]);
	fprintf(stderr, _("          Scans computers 172.17.5.1 to 172.17.5.127 and 127.17.6.1 to\n          172.17.6.127.\n"));
	fprintf(stderr, "\n");
	fprintf(stderr, _("          %s -w MATRIX -u joey,secret\n"), argv[0]);
	fprintf(stderr, _("          Scans computers in workgroup MATRIX and uses 'joey' as username and\n          'secret' as the password for the shares.\n"));
	fprintf(stderr, "\n");
	fprintf(stderr, _("This is SambaScanner %s, released under the GNU GPL-2.\n"), VERSION);
}

int Process_Parameters(int argc, char **argv, StringList *Hostlist) {
	char *Equivalent_Option = NULL;

	/* Default options */
	Options.Workgroup[0] = 0;
	Options.Timings = 1.0;
	Options.DebugLvl = 1;
	Options.ThreadCount = 100;
	Options.PreserveTemp = 0;
	Options.TryPasswords = 1;
	Options.Filter_PrintShares = 1;
	Init_StringList(&Options.UserPass);
	strcpy(Options.SmbClient_Executable, DEFAULT_SMBCLIENT);
	strcpy(Options.NmbLookup_Executable, DEFAULT_NMBLOOKUP);
	strcpy(Options.SmbLister_Executable, DEFAULT_SMBLISTER);
	strcpy(Options.Outfile, "Filelist.txt");

	while (1) {
		int Result;
		int Option_Index = 0;
		static struct option Long_Options[] = {
			{"time", 1, 0, 0},
			{"workgroup", 1, 0, 0},
			{"help", 0, 0, 0},
			{"debug", 1, 0, 0},
			{"smbclient", 1, 0, 0},
			{"nmblookup", 1, 0, 0},
			{"smblister", 1, 0, 0},
			{"outfile", 1, 0, 0},
			{"computer", 1, 0, 0},
			{"ip", 1, 0, 0},
			{"threadcnt", 1, 0, 0},
			{"preserve", 0, 0, 0},
			{"no-passwd", 0, 0, 0},
			{"userpass", 1, 0, 0},
			{"nofilter", 0, 0, 0},
			{0, 0, 0, 0}
		};
		const char *Short_Options = "t:w:hd:s:n:l:o:c:i:r:pau:f";
		Result = getopt_long (argc, argv, Short_Options, Long_Options, &Option_Index);
		if (Result == -1) break;

		if (Result == 0) {
			int i, j;
			if (Equivalent_Option == NULL) {
				Equivalent_Option = (char*)malloc(strlen(Short_Options)+1);
				if (!Equivalent_Option) {
					perror("malloc Equivalent_Option");
					exit(EXIT_FAILURE);
				}
				
				j = 0;
				for (i = 0; i < strlen(Short_Options); i++) {
					if (Short_Options[i] != ':') {
						Equivalent_Option[j] = Short_Options[i];
						j++;
					}
				}
				Equivalent_Option[j] = 0;
			}
			Result = Equivalent_Option[Option_Index];
		}
		
		switch (Result) {
			case 't': {
				double Timeout;
				Timeout = atof(optarg);
				if ((Timeout < TIMEOUT_MIN) || (Timeout > TIMEOUT_MAX)) {
					fprintf(stderr, _("Error: Timeout must be inbetween %.1f and %.1f.\n"), TIMEOUT_MIN, TIMEOUT_MAX);
					return 0;
				}
				Options.Timings=Timeout;
				break;
			}
					  
			case 'w': {
				strncpy(Options.Workgroup, optarg, sizeof(Options.Workgroup));
				Options.Workgroup[sizeof(Options.Workgroup)-1] = '\0';
				break;
			}

			case 'd': {
				int DebugLvl;
				DebugLvl=atoi(optarg);
				if ((DebugLvl < 0) || (DebugLvl > 6)) {
					fprintf(stderr, _("Error: Debuglevel must be inbetween 0 and 6.\n"));
					return 0;
				}
				Options.DebugLvl=DebugLvl;
				break;
			}
			
			case 's': { /* --smbclient */
				strncpy(Options.SmbClient_Executable, optarg, sizeof(Options.SmbClient_Executable));
				Options.SmbClient_Executable[sizeof(Options.SmbClient_Executable) - 1] = 0;
				break;
			}
			case 'n': { /* --nmblookup */
				strncpy(Options.NmbLookup_Executable, optarg, sizeof(Options.NmbLookup_Executable));
				Options.NmbLookup_Executable[sizeof(Options.NmbLookup_Executable) - 1] = 0;
				break;
			}
			case 'l': { /* --smblister */
				strncpy(Options.SmbLister_Executable, optarg, sizeof(Options.SmbLister_Executable));
				Options.SmbLister_Executable[sizeof(Options.SmbLister_Executable) - 1] = 0;
				break;
			}
			case 'o': {	/* --outfile */
				strncpy(Options.Outfile, optarg, sizeof(Options.Outfile));
				Options.Outfile[sizeof(Options.Outfile)-1] = '\0';
				break;
			}
			case 'c': { /* --computer */
				Add_String(Hostlist, optarg);
				break;
			}
			case 'i': { /* --ip */
				int Type = New_IP_Pattern(optarg);
				if (Type) {
					fprintf(stderr, _("Error: Unable to parse IP pattern '%s'.\n"), Pattern_Type_To_String(Type));
					return 0;
				}
				break;
			}
			case 'r': { /* --threadcnt */
				int ThreadCount;
				ThreadCount = atoi(optarg);
				if ((ThreadCount < 1) || (ThreadCount > 100)) {
					fprintf(stderr, _("Error: Threadcount must stay inbetween 1 and 100\n"));
					return 0;
				}
				Options.ThreadCount = ThreadCount;
				break;
			}

			case 'p': { /* --preserve */
				Options.PreserveTemp = 1;
				break;
			}

			case 'a': { /* --passwords */
				Options.TryPasswords = 0;
				break;
			}
			case 'u': { /* --userpass */
				char *Tmp = (char*)malloc(strlen(optarg) + 1);
				char *Token;
				strcpy(Tmp, optarg);
				Token = strtok(Tmp, ",");
				if (Token == NULL) {
					fprintf(stderr, _("Error: There was no username supplied for the '-u'-Option\n"));
					return 0;
				}
				Token = strtok(NULL, ",");
				if (Token == NULL) {
					fprintf(stderr, _("Error: There was no password supplied for the '-u'-Option\n"));
					return 0;
				}
				free(Tmp);

				/* Valid username/password combination -> add to stringlist */
				Add_String_Unique(&Options.UserPass, optarg);
				break;
			}
			
			case 'f': { /* --nofilter */
				Options.Filter_PrintShares = 0;
				break;
			}
					  
			case '?':
			case 'h': {
				Display_Help(argv);
				return 0;
				break;
			}

			default: {
				fprintf(stderr, _("Returned character code was 0x%x\n"), Result);
				Display_Help(argv);
				return 0;
				break;
			}
		}
	}

	if (optind < argc) {
		fprintf(stderr, _("Invalid argument: "));
		while (optind < argc) fprintf(stderr, "%s ", argv[optind++]);
		fprintf(stderr, "\n");
		Display_Help(argv);
		return 0;
	}

	if (Equivalent_Option) {
		free(Equivalent_Option);
	}
	return 1;
}

