/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <libintl.h>

#include "sambascanner.h"
#include "Tools.h"
#include "IP_Range.h"
#include "Processes.h"
#include "String.h"
#include "Parser.h"
#include "Mutex.h"
#include "Threads.h"
#include "IP_Hostname_Pairings.h"

extern struct Options_Structure Options;

int GetUniqueThreadID() {
	static pthread_mutex_t LockUniqueID = PTHREAD_MUTEX_INITIALIZER;
	static int UniqueID = 0;
	int ReturnID;
	pthread_mutex_lock(&LockUniqueID);
	UniqueID++;
	ReturnID = UniqueID;
	pthread_mutex_unlock(&LockUniqueID);
	return ReturnID;
}

void* Redirect_Thread(void *Parameters) {
	struct Redirect_Thread_InputData *Input = (struct Redirect_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	int Read;
	int Cnt = 0;
	int TotalRead = 0;
	int TotalWritten = 0;

	pthread_mutex_lock(&Input->Active);

	message(4, _("Redirect_Thread got unlock for reading/writing, [ProcID %d] [UniqTID %d]\n"), Input->ProcID, Input->UniqueThreadID);
	do {
		int Error;
		Cnt++;
		errno = 0;
		Read = read(Input->Read_FD, Buffer, sizeof(Buffer));
		Error = errno;
		message(6, _("[UniqTID %d] Read #%d (after %d read, %d written): Read = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
		if (Read < 0) {
			message(5, _("[UniqTID %d] Invalid read #%d (after %d read, %d written): Read = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
		} else {
			TotalRead += Read;
		}
		if (Read > 0) {
			int Written;

			errno = 0;
			Written = write(Input->Outfile_FD, Buffer, Read);
			Error = errno;
			message(6, _("[UniqTID %d] Write #%d (after %d read, %d written): Written = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
			if (Written < 0) {
				message(5, _("[UniqTID %d] Invalid write #%d (after %d read, %d written): Written = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
			} else {
				TotalWritten += Written;
			}
			ProcessTable_UpdateEntry_Times(Input->ProcID, Input->TermTime, Input->KillTime);
		}
	} while (Read > 0);
	message(4, _("[UniqTID %d] FD %d -> Redirect_Thread -> FD %d Quit.\n"), Input->UniqueThreadID, Input->Read_FD, Input->Outfile_FD);
	close(Input->Read_FD);
	close(Input->Outfile_FD);

	pthread_mutex_unlock(&Input->Active);
	pthread_mutex_destroy(&Input->Active);

	ProcessTable_Set_Forwarding_Closed(Input->ProcID);

	free(Input);
	pthread_exit(NULL);
	return NULL;
}

void* Process_Watcher_Thread(void *Parameters) {
	struct Process_Watcher_Thread_InputData *Input = (struct Process_Watcher_Thread_InputData*)Parameters;

	while (Input->Running) {
		double Time = GetTime();
		ProcessTable_Rewind();
		while (ProcessTable_Next() != PROCESSTABLE_END) {
			int State;
			if (!ProcessTable_GetCurrent_Running()) continue;

			State = ProcessTable_GetCurrent_SignallingState();
			switch (State) {
				case 0: {
					if (Time > ProcessTable_GetCurrent_TermTime()) {
						pid_t PID = ProcessTable_GetCurrent_PID();
						message(3, _("=> SIGTERM [PID %d], t = %.0f, KillTime = %.0f\n"), PID, Time, ProcessTable_GetCurrent_TermTime());
						kill(PID, SIGTERM);
						ProcessTable_SetCurrent_SignallingState(1);
					}
					break;
				}
				case 1: {
					if (Time > ProcessTable_GetCurrent_KillTime()) {
						pid_t PID = ProcessTable_GetCurrent_PID();
						message(3, _("=> SIGKILL [PID %d], t = %.0f, KillTime = %.0f\n"), PID, Time, ProcessTable_GetCurrent_TermTime());
						kill(PID, SIGKILL);
						ProcessTable_SetCurrent_SignallingState(2);
					}
					break;
				}
			}
		}
		usleep(250 * 1000);		/* 250ms */
	}
	Input->Running = -1;
	pthread_exit(NULL);
	return NULL;
}

void* Process_Host_Thread(void *Parameters) {
	struct Process_Host_Thread_InputData *Input = (struct Process_Host_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_hostXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, _("Couldn't create temporary file in Process_Host_Thread: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}

	message(2, _("Processing host: '%s'\n"), Input->Hostname);
	snprintf(Buffer, sizeof(Buffer), "-g -N -L //%s", Input->Hostname);
	ProcID = Exec_Timeout(Options.SmbClient_Executable, Buffer, TIMEOUT_PROCESSHOST_TERM, TIMEOUT_PROCESSHOST_KILL, fd);
	Result = WaitFor(ProcID);

	pthread_mutex_lock(Input->NumberMutex);
	if (Result < 0) {
		/* Process was killed */
		(*Input->Hostlookup_Processes_Failed)++;
	} else {
		(*Input->Hostlookup_Processes_Succeeded)++;
	}
	pthread_mutex_unlock(Input->NumberMutex);

	Parse_Hostinfo(Input->Hostname, FileBuffer, Input->Hostlist, Input->Sharelist);
	if (!Input->PreserveTemp) unlink(FileBuffer);
	
	free(Input);
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

void* Process_Share_Thread(void *Parameters) {
	struct Process_Share_Thread_InputData *Input = (struct Process_Share_Thread_InputData*)Parameters;
	char CmdParameters[BUFSIZE_SHORT];
	char Buffer[BUFSIZE_SHORT];
	char Buffer2[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_sharesXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, _("Couldn't create temporary file in Process_Share_Thread: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}
	Add_String_Atomic(Input->Resultfilelist, FileBuffer);
	
	message(2, _("Now revealing content of smb:%s with unique thread ID %d\n"), Input->Sharename, Input->UniqueThreadID);

	if (Options.TryPasswords) {
		CmdParameters[0] = 0;
	} else {
		strcpy(CmdParameters, "-g ");
	}
	if (Options.UserPass.Entries > 0) {
		int i;
		for (i = 0; i < Options.UserPass.Entries; i++) {
			snprintf(Buffer, sizeof(Buffer), "%s-u %s ", CmdParameters, Options.UserPass.Entry[i]);
			strcpy(CmdParameters, Buffer);
		}
	}

	Escape_Apostrophes(Buffer2, Input->Sharename, sizeof(Buffer2));
	snprintf(Buffer, sizeof(Buffer), "%s'smb:%s'", CmdParameters, Buffer2);
	strcpy(CmdParameters, Buffer);

	ProcID = Exec_Timeout(Options.SmbLister_Executable, CmdParameters, TIMEOUT_SMBLISTER_TERM, TIMEOUT_SMBLISTER_KILL, fd);
	Result = WaitFor(ProcID);

	if (Result >= 0) {
		if (Result == 0) {
			message(3, _("smblister returned on [ProcID %d] successfully.\n"), ProcID);
		} else {
			message(2, _("smblister returned on [ProcID %d] with error: %s\n"), ProcID, GetSmblisterError(Result));
		}
	}
	
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

void *Process_IP_Address_Thread(void *Parameters) {
	struct Process_IP_Address_Thread_InputData *Input = (struct Process_IP_Address_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	message(3, _("Now processing IP %d.%d.%d.%d\n"), IP_ADDRESS(Input->IP_Address));

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_ipaddrXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, _("Couldn't create temporary file in Process_IP_Address_Thread: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	snprintf(Buffer, sizeof(Buffer), "-A %d.%d.%d.%d", IP_ADDRESS(Input->IP_Address));
	ProcID = Exec_Timeout(Options.NmbLookup_Executable, Buffer, TIMEOUT_IPSCAN_TERM, TIMEOUT_IPSCAN_KILL, fd);
	Result = WaitFor(ProcID);

	Parse_IP_Address_Info(FileBuffer, Input->Hostlist, Buffer, sizeof(Buffer));
	if (Buffer[0] != 0) {
		Add_IP_Hostname_Pairing(Input->IP_Address, Buffer);
	}
	if (!Input->PreserveTemp) unlink(FileBuffer);

	free(Parameters);
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

int pthread_create_detached(pthread_t *thread, void *(*start_routine)(void*), void *arg) {
	pthread_attr_t Attrs;
	pthread_attr_init(&Attrs);
	pthread_attr_setdetachstate(&Attrs, PTHREAD_CREATE_DETACHED);
	return pthread_create(thread, &Attrs, start_routine, arg);
}

